use color_eyre::Result;
use crossterm::event::{KeyCode, KeyEvent};
use ratatui::{
    layout::{Constraint, Flex, Layout, Rect},
    Frame,
};

use std::sync::mpsc::Sender;
use tokio::sync::mpsc::UnboundedSender;

mod balance;
mod transfer;
mod address_book;
mod add_account;
mod rename_account;
mod event_logs;
mod accounts;
mod overview;
mod add_address_book_record;
mod rename_address_book_record;
mod account_details;
mod staking_ledger;
mod bond_popup;
mod payee_popup;
mod current_validators;
mod current_validator_details;
mod rename_known_validator;

use balance::Balance;
use transfer::Transfer;
use address_book::AddressBook;
use add_account::AddAccount;
use rename_account::RenameAccount;
use event_logs::EventLogs;
use accounts::Accounts;
use overview::Overview;
use add_address_book_record::AddAddressBookRecord;
use rename_address_book_record::RenameAddressBookRecord;
use account_details::AccountDetails;
use staking_ledger::StakingLedger;
use bond_popup::BondPopup;
use payee_popup::PayeePopup;
use current_validators::CurrentValidators;
use current_validator_details::CurrentValidatorDetails;
use rename_known_validator::RenameKnownValidator;

use super::Component;
use crate::{action::Action, app::Mode, config::Config};

#[derive(Debug, Copy, Clone, PartialEq)]
pub enum CurrentTab {
    Nothing,
    Accounts,
    AddressBook,
    EventLogs,
    AddAccount,
    AddAddressBookRecord,
    RenameAccount,
    RenameAddressBookRecord,
    Transfer,
    AccountDetails,
    BondPopup,
    PayeePopup,
    CurrentValidatorsPopup,
}

pub trait PartialComponent: Component {
    fn set_active(&mut self, current_tab: CurrentTab);
}

pub struct Wallet {
    is_active: bool,
    current_tab: CurrentTab,
    previous_tab: CurrentTab,
    components: Vec<Box<dyn PartialComponent>>,
}

impl Default for Wallet {
    fn default() -> Self {
        Self {
            is_active: false,
            current_tab: CurrentTab::Accounts,
            previous_tab: CurrentTab::Accounts,
            components: vec![
                Box::new(Overview::default()),
                Box::new(Accounts::default()),
                Box::new(Balance::default()),
                Box::new(StakingLedger::default()),
                Box::new(AddressBook::default()),
                Box::new(EventLogs::default()),
                Box::new(AddAccount::default()),
                Box::new(RenameAccount::default()),
                Box::new(AddAddressBookRecord::default()),
                Box::new(RenameAddressBookRecord::default()),
                Box::new(Transfer::default()),
                Box::new(AccountDetails::default()),
                Box::new(BondPopup::default()),
                Box::new(PayeePopup::default()),
                Box::new(CurrentValidators::default()),
                Box::new(CurrentValidatorDetails::default()),
                Box::new(RenameKnownValidator::default()),
            ],
        }
    }
}

impl Wallet {
    fn move_left(&mut self) {
        match self.current_tab {
            CurrentTab::EventLogs => self.current_tab = CurrentTab::AddressBook,
            CurrentTab::AddressBook => self.current_tab = CurrentTab::Accounts,
            _ => {}
        }
    }

    fn move_right(&mut self) {
        match self.current_tab {
            CurrentTab::Nothing => self.current_tab = CurrentTab::Accounts,
            CurrentTab::Accounts => self.current_tab = CurrentTab::AddressBook,
            CurrentTab::AddressBook => self.current_tab = CurrentTab::EventLogs,
            _ => {}
        }
    }
}

impl Component for Wallet {
    fn register_network_handler(&mut self, tx: Sender<Action>) -> Result<()> {
        for component in self.components.iter_mut() {
            component.register_network_handler(tx.clone())?;
        }
        Ok(())
    }

    fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
        for component in self.components.iter_mut() {
            component.register_action_handler(tx.clone())?;
        }
        Ok(())
    }

    fn register_config_handler(&mut self, config: Config) -> Result<()> {
        for component in self.components.iter_mut() {
            component.register_config_handler(config.clone())?;
        }
        Ok(())
    }

    fn handle_key_event(&mut self, key: KeyEvent) -> Result<Option<Action>> {
        if !self.is_active { return Ok(None) }

        match self.current_tab {
            // block the default key handle for popups
            CurrentTab::AddAccount | 
                CurrentTab::RenameAccount | 
                CurrentTab::RenameAddressBookRecord | 
                CurrentTab::Transfer |
                CurrentTab::AccountDetails |
                CurrentTab::BondPopup |
                CurrentTab::PayeePopup |
                CurrentTab::CurrentValidatorsPopup |
                CurrentTab::AddAddressBookRecord => {
                    for component in self.components.iter_mut() {
                        component.handle_key_event(key)?;
                    }
                }, 
            _ => match key.code {
                KeyCode::Esc => {
                    self.is_active = false;
                    self.current_tab = CurrentTab::Nothing;
                    return Ok(Some(Action::SetActiveScreen(Mode::Menu)));
                },
                KeyCode::Char('W') => {
                    self.previous_tab = self.current_tab;
                    self.current_tab = CurrentTab::AddAccount;
                },
                KeyCode::Char('A') => {
                    self.previous_tab = self.current_tab;
                    self.current_tab = CurrentTab::AddAddressBookRecord;
                },
                KeyCode::Char('T') => {
                    self.previous_tab = self.current_tab;
                    self.current_tab = CurrentTab::Transfer;
                },
                KeyCode::Char('B') => {
                    self.previous_tab = self.current_tab;
                    self.current_tab = CurrentTab::BondPopup;
                },
                KeyCode::Char('I') => {
                    self.previous_tab = self.current_tab;
                    self.current_tab = CurrentTab::PayeePopup;
                },
                KeyCode::Char('N') => {
                    self.previous_tab = self.current_tab;
                    self.current_tab = CurrentTab::CurrentValidatorsPopup;
                },
                KeyCode::Char('l') | KeyCode::Right => self.move_right(),
                KeyCode::Char('h') | KeyCode::Left => self.move_left(),
                _ => {
                    for component in self.components.iter_mut() {
                        component.handle_key_event(key)?;
                    }
                },
            }
        }
        Ok(None)
    }

    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        match action {
            Action::SetActiveScreen(Mode::Wallet) => {
                self.is_active = true;
                self.current_tab = CurrentTab::Accounts;
                self.previous_tab = CurrentTab::Accounts;
            },
            Action::RenameAccount(_) => {
                self.previous_tab = self.current_tab;
                self.current_tab = CurrentTab::RenameAccount;
            },
            Action::RenameAddressBookRecord(_) => {
                self.previous_tab = self.current_tab;
                self.current_tab = CurrentTab::RenameAddressBookRecord;
            }
            Action::TransferTo(_) => {
                self.previous_tab = self.current_tab;
                self.current_tab = CurrentTab::Transfer;
            },
            Action::AccountDetailsOf(_, _) => {
                self.previous_tab = self.current_tab;
                self.current_tab = CurrentTab::AccountDetails;
            },
            Action::ClosePopup => self.current_tab = self.previous_tab,
            _ => {}
        }
        for component in self.components.iter_mut() {
            component.set_active(self.current_tab.clone());
            component.update(action.clone())?;
        }
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let screen = super::screen_layout(area);
        for component in self.components.iter_mut() {
            component.draw(frame, screen)?;
        }
        Ok(())
    }
}

pub fn wallet_layout(area: Rect) -> [Rect; 2] {
    Layout::vertical([
        Constraint::Percentage(75),
        Constraint::Percentage(25),
    ]).areas(area)
}

pub fn bars_layout(area: Rect) -> [Rect; 2] {
    let [place, _] = wallet_layout(area);
    Layout::horizontal([
        Constraint::Percentage(50),
        Constraint::Percentage(50),
    ]).areas(place)
}

pub fn account_layout(area: Rect) -> [Rect; 4] {
    let [place, _] = bars_layout(area);
    Layout::vertical([
        Constraint::Max(4),
        Constraint::Min(0),
        Constraint::Max(7),
        Constraint::Max(6),
    ]).areas(place)
}

pub fn nominator_layout(area: Rect) -> [Rect; 2] {
    let v = Layout::vertical([Constraint::Max(11)]).flex(Flex::Center);
    let [area] = v.areas(area);
    Layout::horizontal([
        Constraint::Percentage(60),
        Constraint::Percentage(40),
    ]).areas(area)
}
